// Copyright (C) 2021 by the IntelliStream team
// (https://github.com/intellistream)

//
// Created by Shuhao Zhang on 19/07/2021.
//

#ifndef SESAME_INCLUDE_ENGINE_ENGINE_HPP_
#define SESAME_INCLUDE_ENGINE_ENGINE_HPP_

namespace SESAME {
class Engine {};
} // namespace SESAME
#endif // SESAME_INCLUDE_ENGINE_ENGINE_HPP_
